Package: quickSentiment
Title: A Fast and Flexible Pipeline for Text Classification
Version: 0.3.4
Authors@R: 
    person("Alabhya", "Dahal", email = "alabhya.dahal@gmail.com", role = c("aut", "cre"))
Description: A high-level pipeline that simplifies text classification into three streamlined steps: 
    preprocessing, model training, and standardized prediction. 
    It unifies the interface for multiple algorithms (including 'glmnet', 'ranger', 
    'xgboost', and 'naivebayes') and memory-efficient sparse matrix vectorization 
    methods (Bag-of-Words, Term Frequency, TF-IDF, and Binary). Users can go from 
    raw text to a fully evaluated sentiment model, complete with ROC-optimized 
    thresholds, in just a few function calls. The resulting model artifact 
    automatically aligns the vocabulary of new datasets during the prediction phase, 
    safely appending predicted classes and probability matrices directly to the 
    user's original dataframe to preserve metadata.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: doParallel, foreach, glmnet, magrittr, Matrix, methods,
        naivebayes, quanteda, ranger, stopwords, stringr, textstem,
        xgboost
VignetteBuilder: knitr
Suggests: knitr, rmarkdown, spelling
Language: en-US
NeedsCompilation: no
Packaged: 2026-04-17 01:50:05 UTC; meala
Author: Alabhya Dahal [aut, cre]
Maintainer: Alabhya Dahal <alabhya.dahal@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-17 03:30:02 UTC
Built: R 4.6.0; ; 2026-04-30 23:04:19 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,quickSentiment_prc)
S3method(plot,quickSentiment_roc)
S3method(print,quickSentiment_eval)
export(BOW_test)
export(BOW_train)
export(evaluate_performance)
export(logit_model)
export(nb_model)
export(pipeline)
export(pre_process)
export(predict_sentiment)
export(qs_negations)
export(rf_model)
export(xgb_model)
importFrom(Matrix,Diagonal)
importFrom(doParallel,registerDoParallel)
importFrom(foreach,registerDoSEQ)
importFrom(glmnet,cv.glmnet)
importFrom(graphics,abline)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(magrittr,"%>%")
importFrom(methods,as)
importFrom(naivebayes,multinomial_naive_bayes)
importFrom(parallel,detectCores)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(quanteda,dfm)
importFrom(quanteda,dfm_match)
importFrom(quanteda,dfm_tfidf)
importFrom(quanteda,dfm_weight)
importFrom(quanteda,docfreq)
importFrom(quanteda,featnames)
importFrom(quanteda,ndoc)
importFrom(quanteda,tokens)
importFrom(quanteda,tokens_ngrams)
importFrom(quanteda,tokens_select)
importFrom(ranger,ranger)
importFrom(stats,predict)
importFrom(stopwords,stopwords)
importFrom(stringr,str_remove_all)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_squish)
importFrom(textstem,lemmatize_strings)
importFrom(xgboost,xgb.DMatrix)
importFrom(xgboost,xgb.cv)
importFrom(xgboost,xgb.train)

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/quickSentiment)](https://cran.r-project.org/package=quickSentiment)
[![CRAN Downloads](https://cranlogs.r-pkg.org/badges/grand-total/quickSentiment)](https://cran.r-project.org/package=quickSentiment)
[![CRAN checks](https://badges.cranchecks.info/summary/quickSentiment.svg)](https://cran.r-project.org/web/checks/check_results_quickSentiment.html)
<img src="https://img.shields.io/badge/R-276DC3?style=flat-square&logo=r&logoColor=white" />


[PLEASE REFER TO DOCUMENTATION IN CRAN OR DOWNLOAD THE PACKAGE TO VIEW DOCUMENTATION FOR MORE DETAILS]

quickSentiment: A Fast and Flexible Pipeline for Text Classification in R
quickSentiment is an R package designed to streamline the process of text classification. It provides a complete, end-to-end workflow from text cleaning to model training, evaluation, and prediction. Built on a modular architecture, it allows users to easily experiment with different vectorization methods and high-performance machine learning models.

Key Features
Modular Pipeline: A single, powerful pipeline() function to run the entire training and evaluation process.

Multiple Vectorization Methods: Includes support for Bag-of-Words, Term Frequency (TF), and TF-IDF, with easy n-gram integration.

High-Performance Models: Comes with built-in support for three powerful and popular classification models:

Logistic Regression (via glmnet for speed and regularization)

Random Forest (via the high-performance ranger package)

XGBoost (via the industry-standard xgboost package)

Naive Bayesian ( via naivebayes package)

Reproducible Predictions: The pipeline returns a self-contained artifact object, bundling the trained model and all necessary components to ensure that predictions on new data are consistent and reliable.

Installation
You can install the development version of quickSentiment from GitHub with:

```bash
# install.packages("devtools")
devtools::install_github("AlabhyaMe/quickSentiment")
# UPDATES: The package is now live in CRAN. You can download it directly in RStudio, or
install.package("quickSentiment")

```
Core Workflow: The Three Main Functions
The quickSentiment workflow is designed to be logical and flexible, centered around three key functions.

1. pre_process_final(): Cleaning Your Text
The first step is to clean your raw text data. This function is kept separate from the main pipeline to allow you to use the same cleaned text for multiple different tasks in the future (e.g., classification, topic modeling).

It handles a variety of common cleaning steps, including lowercasing, lemmatization, and removing URLs, HTML, and stopwords.

Usage:
```bash
library(quickSentiment)
library(readr)

# Load your data
my_data <- read_csv("path/to/your/data.csv")

# Create a new column of cleaned text
my_data$cleaned_text <- pre_process(my_data$reviewText)
```

2. pipeline(): The Main Engine for Training
This is the core function of the package. It takes your preprocessed data and handles everything else: splitting the data, vectorizing the text, training a model, and evaluating its performance.

You can easily swap out vectorization methods or models just by changing a string argument.

Usage:
```bash
# Train a TF-IDF Logistic Regression model with bigrams
pipeline_artifacts <- pipeline(
  text_vector = df$cleaned_text,
  sentiment_vector = df$rating, # The original column with ratings or labels
  vect_method = "tfidf",
  model_name = "logit",
  n_gram = 2
)
```
The function will print a detailed evaluation report from caret and return a list containing all the necessary "artifacts" for prediction.

3. prediction(): Scoring New, Unseen Data
Once your pipeline has run, you can use the pipeline_artifacts object it returned to make predictions on new data. This function is generic and automatically adapts to the type of model you trained (glmnet, ranger, or xgboost).

Usage:
```bash
# Create a vector of new, raw text to predict on
new_reviews <- "your new data file."
new_rviews$cleaned <- pre_process(new_reviews$"the text column")

# The prediction function uses the artifacts to ensure a consistent workflow
final_predictions <- predict_sentiment(
df$cleaned
pipeline_object = pipeline_artifacts,
)

print(final_predictions)
```
Behind the Hood: How it Works
quickSentiment is designed to be both user-friendly and powerful. This is achieved through a few key design choices:

The "Model Contract": Each model function (logit, rf_model, xgboost_model) follows a strict contract: it must return a list containing the predictions (pred) and the trained model object (model). This allows the main pipeline to treat them interchangeably, making the system highly modular and easy to extend with new models.

High-Performance Backends: The package doesn't reinvent the wheel. It stands on the shoulders of giants by using the best-in-class, high-performance packages for each task:

quanteda for fast and memory-efficient text processing.

glmnet for its incredibly fast, pathwise algorithm for regularized regression.

ranger for a multi-threaded, C++ implementation of Random Forest.

xgboost for its industry-leading speed and predictive accuracy.

Train and test internally, returns ROC, accuracy, precision, recall, specificity, and f1_score for the user.

Customized prediction function that keeps the artifact of the model consistent.  

Self-Contained Artifacts: The pipeline() function returns a single R object that contains everything needed for reproducible predictions: the trained model, the DFM vocabulary template, the preprocessing function, and the n-gram settings. This "all-in-one" object prevents common errors and ensures that new data is always processed in the exact same way as the training data.

Future Development
This package is stable and mature

This package is licensed under the MIT License.

## ----include = FALSE----------------------------------------------------------
library(quickSentiment)


## ----setup--------------------------------------------------------------------
library(doParallel)
# CRAN limits the number of cores used during package checks
cores <- min(2, parallel::detectCores())
registerDoParallel(cores = cores)

## -----------------------------------------------------------------------------
# Look for the file in the installed package first
csv_path <- system.file("extdata", "tweets.csv", package = "quickSentiment")

# Fallback for when you are building the package locally
if (csv_path == "") {
  csv_path <- "../inst/extdata/tweets.csv"
}
tweets <- read.csv(csv_path)
set.seed(123)

## -----------------------------------------------------------------------------
tweets$cleaned_text <- pre_process(tweets$Tweet)
tweets$sentiment = ifelse(tweets$Avg>0,'P','N')

## -----------------------------------------------------------------------------
result <- pipeline(
  # --- Define the vectorization method ---
  # Options: "bow" (raw counts), "tf" (term frequency), "tfidf", "binary"
  vect_method = "tf",
  
  # --- Define the model to train ---
  # Options: "logit", "rf", "xgb","nb"
  model_name = "rf",
  
  # --- Specify the data and column names ---
  text_vector = tweets$cleaned_text  ,   # The column with our preprocessed text
  sentiment_vector = tweets$sentiment,    # The column with the target variable
  
  # --- Set vectorization options ---
  # Use n_gram = 2 for unigrams + bigrams, or 1 for just unigrams
  n_gram = 1,
  parallel = cores
)

## -----------------------------------------------------------------------------
evaluate_result<- evaluate_performance(result$probs[,2],result$y_test,"P")
evaluate_result

## -----------------------------------------------------------------------------
plot(evaluate_result$roc)

## -----------------------------------------------------------------------------
plot(evaluate_result$prc)

## -----------------------------------------------------------------------------

predicted_tweets <- predict_sentiment(
  pipeline_object = result,
  tweets$cleaned_text
)
head(predicted_tweets)

#BOW.R
#' Train a Bag-of-Words Model
#' @param doc A character vector of documents to be processed.
#' @param weighting_scheme A string specifying the weighting to apply.
#'   Defaults to \code{"bag_of_words"}.
#'   \itemize{
#'     \item \code{"bag_of_words"} (Alias: \code{"bow"}) - Standard count of words.
#'     \item \code{"term_frequency"} (Alias: \code{"tf"}) - Normalized counts (frequency relative to document length).
#'     \item \code{"tfidf"} (Alias: \code{"tf-idf"}) - Term Frequency-Inverse Document Frequency.
#'     \item \code{"binary"} - Presence/Absence (1/0).
#'   }
#' @param ngram_size An integer specifying the maximum n-gram size. For example,
#' `ngram_size = 1` will create unigrams only; `ngram_size = 2` will create unigrams and bigrams. Defaults to 1.
#' @return An object of class \code{"qs_bow_fit"} containing:
#'   \itemize{
#'     \item \code{dfm_template}: a quanteda \code{dfm} template
#'     \item \code{weighting_scheme}: the weighting used
#'     \item \code{ngram_size}: the n-gram size used
#'   }#'
#' @importFrom quanteda tokens tokens_ngrams tokens_select dfm dfm_weight ndoc docfreq dfm_tfidf
#' @importFrom magrittr %>%
#' @importFrom stopwords stopwords
#' @export
#' @examples
#' txt <- c("text one", "text two text")
#' fit <- BOW_train(txt, weighting_scheme = "bow")
#' fit$dfm_template
#'
BOW_train <- function(doc,weighting_scheme = "bow",ngram_size=1) {
  idf_vector <- NULL

  #check for proper inputs
  weighting_scheme <- tolower(trimws(weighting_scheme))
  if (weighting_scheme == "bow") weighting_scheme <- "bag_of_words"
  if (weighting_scheme == "tf")  weighting_scheme <- "term_frequency"
  if (weighting_scheme == "tf-idf")  weighting_scheme <- "tfidf"

  allowed <- c("bag_of_words", "binary", "term_frequency", "tfidf")

  if (!weighting_scheme %in% allowed) {
    stop(sprintf("Invalid weighting_scheme '%s'. Must be one of: %s",
                 weighting_scheme, paste(allowed, collapse = ", ")),
         call. = FALSE)
  }

  #N-Gram check
  ngram_size <- as.integer(ngram_size)
  if (ngram_size < 1) {
    stop("`ngram_size` must be >= 1.", call. = FALSE)
  }

  message(paste0("  - Fitting BoW model (", weighting_scheme, ") on training data..."))

  # Create the DFM
  dfm_raw <- doc %>%
    quanteda::tokens() %>%
    quanteda::tokens_ngrams(n = 1:ngram_size) %>%
    quanteda::dfm()

  # Store IDF vector separately ONLY if using tfidf for manual matching in test
  # Note: quanteda's dfm_tfidf is preferred, but for BOW_test to apply training
  # weights to new data, we still need the training IDF values.
  idf_vals <- NULL

  if (weighting_scheme == "tfidf") {
    # Using the standard smoothed IDF formula as you had before
    N <- quanteda::ndoc(dfm_raw)
    df_counts <- quanteda::docfreq(dfm_raw)
    idf_vals <- log((N + 1) / (df_counts + 1)) + 1
  }
  # Weighting Logic
  dfm_final <- switch(weighting_scheme,
                      "binary" = quanteda::dfm_weight(dfm_raw, scheme = "boolean"),
                      "term_frequency"     = quanteda::dfm_weight(dfm_raw, scheme = "prop"),
                      "tfidf"  = quanteda::dfm_tfidf(dfm_raw, scheme_tf = "prop", scheme_df = "inverse",base=exp(1)), #force natural log,
                      dfm_raw # default bow
  )

  #returns not just the dfm but also the weighting scheme and ngram size for future reference

  fit <- list(
    dfm_template = dfm_final,
    weighting_scheme = weighting_scheme,
    ngram_size = ngram_size,
    idf_vector = if (weighting_scheme == "tfidf") idf_vals else NULL
  )
  class(fit) <- "qs_bow_fit"
  return(fit)
}
#' Transform New Text into a Document-Feature Matrix
#'
#' This function takes a character vector of new documents and transforms it
#' into a DFM that has the exact same features as a pre-fitted training DFM,
#' ensuring consistency for prediction.
#'
#' @param doc A character vector of new documents to be processed.
#' @param fit A fitted BoW object returned by \code{BOW_train()}.
#' @return A quanteda \code{dfm} aligned to the training features.
#'
#' @importFrom quanteda tokens tokens_ngrams tokens_select dfm dfm_match featnames dfm_weight
#' @importFrom magrittr %>%
#' @importFrom Matrix Diagonal
#' @export
#' @examples
#' train_txt <- c("apple orange banana", "apple apple")
#' fit <- BOW_train(train_txt, weighting_scheme = "bow")
#' new_txt <- c("banana pear", "orange apple")
#' test_dfm <- BOW_test(new_txt, fit)
#' test_dfm

#  BOW_test function

BOW_test <- function(doc, fit) {
  # check for proper inputs
  if (!inherits(fit, "qs_bow_fit")) {
    stop("`fit` must be an object of class 'qs_bow_fit'.", call. = FALSE)
  }

  if (is.null(fit$dfm_template) || is.null(fit$weighting_scheme) || is.null(fit$ngram_size)) {
    stop("`fit` is missing required components.", call. = FALSE)
  }

  # Extract artifacts
  training_dfm <- fit$dfm_template
  weighting_scheme <- fit$weighting_scheme
  ngram_size <- fit$ngram_size
  saved_idf <- fit$idf_vector

  message(paste0("  - Applying BoW transformation (", weighting_scheme, ") to new data..."))

  # Create a raw DFM and match its features to the training template
  dfm_raw <- doc %>%
    quanteda::tokens() %>%
    quanteda::tokens_ngrams(n = 1:ngram_size)%>%
    quanteda::dfm()

  # Align features to training template
  dfm_matched <- quanteda::dfm_match(dfm_raw, features = quanteda::featnames(training_dfm))

  # 3. Apply the SAME weighting scheme
  dfm_final <- switch(fit$weighting_scheme,
                      "binary" = quanteda::dfm_weight(dfm_matched, scheme = "boolean"),
                      "term_frequency"     = quanteda::dfm_weight(dfm_matched, scheme = "prop"),
                      "tfidf"  = {
                        # Use the specific IDF vector calculated during training
                        tf_test <- quanteda::dfm_weight(dfm_matched, scheme = "prop")
                        weighted_mat <- tf_test %*% Matrix::Diagonal(x = fit$idf_vector)
                        quanteda::as.dfm(weighted_mat)
                      },
                      dfm_matched # Default for "bag of words" (raw counts)
  )
  return(dfm_final)
}

#' Train a Regularized Logistic Regression Model using glmnet
#'
#' This function trains a logistic regression model using Lasso regularization
#' via the glmnet package. It uses cross-validation to automatically find the
#' optimal regularization strength (lambda).
#'
#' @param train_vectorized The training feature matrix (e.g., a `dfm` from quanteda).
#'   This should be a sparse matrix.
#' @param Y The response variable for the training set. Should be a factor for
#'   classification.
#' @param test_vectorized The test feature matrix, which must have the same
#'   features as `train_vectorized`.
#' @param parallel Logical
#' @param tune Logical
#'
#' @return A list containing two elements:
#'   \item{pred}{A vector of class predictions for the test set.}
#'   \item{probs}{A matrix of predicted probabilities.}
#'   \item{model}{The final, trained `cv.glmnet` model object.}
#'   \item{best_lambda}{The optimal lambda value found during cross-validation.}
#'
#' @importFrom glmnet cv.glmnet
#' @importFrom stats predict
#' @importFrom doParallel registerDoParallel
#'
#' @export
#' @examples
#' \dontrun{
#' # Create dummy vectorized training and test data
#' train_matrix <- matrix(runif(100), nrow = 10, ncol = 10)
#' test_matrix <- matrix(runif(50), nrow = 5, ncol = 10)
#'
#' # Provide column names (vocabulary) required by glmnet
#' colnames(train_matrix) <- paste0("word", 1:10)
#' colnames(test_matrix) <- paste0("word", 1:10)
#'
#' y_train <- factor(sample(c("P", "N"), 10, replace = TRUE))
#'
#' # Run logistic regression model (glmnet)
#' model_results <- logit_model(train_matrix, y_train, test_matrix)
#' }
logit_model <- function(train_vectorized, Y, test_vectorized, parallel=FALSE, tune = FALSE){
  message("\n--- Running Logistic Regression (logit) Function ---\n")
  fam <- if (nlevels(Y) > 2) "multinomial" else "binomial"
  final_fit <- NULL
  best_s <- NULL

  if (isTRUE(tune)) {
    message("--- Tuning Elastic Net (alpha and lambda) ---\n")
    alpha_grid <- seq(0, 1, by = 0.2)
    best_cv_error <- Inf

    for (a in alpha_grid) {
      # cv.glmnet automatically handles lambda tuning and k-fold CV
      cv_fit <- cv.glmnet(
        x = train_vectorized,
        y = Y,
        family = fam,
        alpha = a,
        nfolds = 5,
        parallel = parallel
      )

    # Get the minimum error for this alpha
    min_err <- min(cv_fit$cvm)
    # If this alpha is better, save the model and parameters
    if (min_err < best_cv_error) {
      best_cv_error <- min_err
      final_fit <- cv_fit
      best_s <- cv_fit$lambda.min
      }

  }
  } else {
  final_fit <- cv.glmnet(
    x = train_vectorized,
    y = Y ,
    alpha = 1,           #  Lasso regularization
    nfolds = 5,
    family= fam,
    parallel = parallel
  )
  best_s <- final_fit$lambda.min
  }

   y_probs <- predict(final_fit,
                    newx = test_vectorized,
                    s = best_s,
                    type = "response")

  if (fam == "binomial") {
    # For binary, we use a standard 0.5 cutoff for the internal report
    p_class2 <- as.vector(y_probs)
    probs_matrix <- cbind(1 - p_class2, p_class2)
    colnames(probs_matrix) <- levels(Y)
    y_pred_idx <- ifelse(p_class2 > 0.5, 2, 1)
  } else {
    # For multi-class, we pick the column with the highest probability
    probs_matrix <- drop(y_probs)
    colnames(probs_matrix) <- levels(Y)

    # Prediction: Pick the column with max probability
    y_pred_idx <- max.col(probs_matrix)
  }
  y_pred_internal <- factor(levels(Y)[y_pred_idx], levels = levels(Y))

  results <- list(
    pred = y_pred_internal, #category
    probs = probs_matrix,
    model = final_fit,
    best_lambda = best_s
  )

  return(results)

}

#' Plot ROC Curve
#'
#' @param x An object of class `quickSentiment_roc`.
#' @param ... Additional graphical parameters.
#' @importFrom graphics par plot points abline
#' @export
plot.quickSentiment_roc <- function(x, ...) {
  # Professional square margins
  oldpar <- par(no.readonly = TRUE)
  on.exit(par(oldpar))
  par(mar = c(4.5, 4.5, 3, 1.5), pty = "s")

  plot(x$curve$FPR, x$curve$TPR_Recall,
       type = "l", col = "darkred", lwd = 2,
       xlab = "False Positive Rate", ylab = "True Positive Rate",
       main = paste("ROC Curve (AUC =", round(x$auc, 3), ")"))

  abline(a = 0, b = 1, lty = 2, col = "gray")

  best_idx <- which(x$curve$Threshold == x$best_threshold)[1]
  points(x$curve$FPR[best_idx], x$curve$TPR_Recall[best_idx],
         col = "blue", pch = 19, cex = 1.5)
}


#' Plot Precision-Recall Curve
#'
#' @param x An object of class `quickSentiment_prc`.
#' @param ... Additional graphical parameters.
#' @importFrom graphics par plot points
#' @export
plot.quickSentiment_prc <- function(x, ...) {
  # Professional square margins
  oldpar <- par(no.readonly = TRUE)
  on.exit(par(oldpar))
  par(mar = c(4.5, 4.5, 3, 1.5), pty = "s")

  plot(x$curve$TPR_Recall, x$curve$Precision,
       type = "l", col = "darkblue", lwd = 2,
       xlab = "Recall (True Positive Rate)", ylab = "Precision",
       main = paste("PR Curve (AUC =", round(x$auc, 3), ")"))

  best_idx <- which(x$curve$Threshold == x$best_threshold)[1]
  points(x$curve$TPR_Recall[best_idx], x$curve$Precision[best_idx],
         col = "red", pch = 19, cex = 1.5)
}
#' Print quickSentiment Evaluation Results
#'
#' @param x An object of class `quickSentiment_eval`.
#' @param ... Further arguments passed to or from other methods.
#' @export
print.quickSentiment_eval <- function(x, ...) {

  cat("=========================================\n")
  cat(" quickSentiment Model Evaluation \n")
  cat("=========================================\n")
  cat("Target Class:  ", x$target_class, "\n\n")

  cat("--- Global Metrics ---\n")
  cat("ROC AUC:       ", round(x$auc_roc, 4), "\n")
  cat("PR AUC:        ", round(x$auc_pr, 4), "\n\n")

  cat("--- Optimal Thresholds ---\n")
  cat("Best ROC Threshold (Youden's J): ", round(x$best_threshold_roc, 4), "\n")
  cat("Best PR Threshold (F1-Score):    ", round(x$best_threshold_pr, 4), "\n")
  cat("Accuracy at Best PR Threshold:   ", round(x$accuracy_at_best, 4), "\n\n")

  cat("--- Threshold Summary Table ---\n")
  # Print the user's custom table, rounded nicely to 3 decimal places
  print(round(x$threshold_summary, 3), row.names = FALSE)

  cat("\n(Note: Use plot() to view ROC and PR curves)\n")

  # Return the object invisibly so the data isn't lost
  invisible(x)
}

#' Multinomial Naive Bayes for Text Classification
#'
#' @param train_vectorized The training feature matrix (e.g., a `dfm` from quanteda).
#'   This should be a sparse matrix.
#' @param Y The response variable for the training set. Should be a factor for
#'   classification.
#' @param test_vectorized The test feature matrix, which must have the same
#'   features as `train_vectorized`
#' @param parallel Logical
#' @param tune Logical. If TRUE, tests different Laplace smoothing values.
#' @return A list containing four elements:
#'   \item{pred}{A vector of class predictions for the test set.}
#'   \item{probs}{A matrix of predicted probabilities.}
#'   \item{model}{The final, trained `naivebayes` model object.}
#'   \item{best_lambda}{Placeholder (NULL) for pipeline consistency.}
#' @importFrom naivebayes multinomial_naive_bayes
#' @importFrom stats predict
#' @importFrom methods as
#' @export
#' @examples
#' # 1. Create dummy numeric matrices with BOTH row and column names
#' train_matrix <- matrix(
#'   as.numeric(sample(0:5, 100, replace = TRUE)),
#'   nrow = 10, ncol = 10,
#'   dimnames = list(paste0("doc", 1:10), paste0("word", 1:10))
#' )
#'
#' test_matrix <- matrix(
#'   as.numeric(sample(0:5, 50, replace = TRUE)),
#'   nrow = 5, ncol = 10,
#'   dimnames = list(paste0("doc", 1:5), paste0("word", 1:10))
#' )
#'
#' # 2. Create dummy target variable
#' y_train <- factor(sample(c("P", "N"), 10, replace = TRUE))
#'
#' # 3. Run model
#' model_results <- nb_model(train_matrix, y_train, test_matrix)
#' print(model_results$pred)

nb_model <- function(train_vectorized, Y, test_vectorized, parallel = FALSE, tune = FALSE) {

  message("\n--- Training Multinomial Naive Bayes ---\n")

  best_laplace <- 1 # Default standard

  # --- MEMORY SAFETY: Cast quanteda DFM to dgCMatrix ---
  X_train_sparse <- methods::as(train_vectorized, "dgCMatrix")
  X_test_sparse <- methods::as(test_vectorized, "dgCMatrix")

  if (isTRUE(tune)) {
    message("  - Tuning: Performing 5-fold CV for Laplace smoothing...")

    # 1. Define Grid
    laplace_grid <- c(0.1, 0.5, 1.0, 1.5, 2.0)

    # 2. Create Folds (stratified)
    n_rows <- nrow(X_train_sparse)
    folds <- split(sample(seq_len(n_rows)), rep(1:5, length.out = n_rows))

    # Store errors for each laplace value
    grid_errors <- numeric(length(laplace_grid))

    # 3. Manual CV Loop (Keeps data sparse!)
    for (i in seq_along(laplace_grid)) {
      L <- laplace_grid[i]
      fold_errors <- numeric(length(folds))

      for (k in seq_along(folds)) {
        # Split data
        val_idx <- folds[[k]]

        # specific for sparse matrix slicing
        X_tr_fold <- X_train_sparse[-val_idx, , drop=FALSE]
        y_tr_fold <- Y[-val_idx]
        X_val_fold <- X_train_sparse[val_idx, , drop=FALSE]
        y_val_fold <- Y[val_idx]

        # Train & Predict
        fit <- naivebayes::multinomial_naive_bayes(x = X_tr_fold, y = y_tr_fold, laplace = L)
        preds <- stats::predict(fit, newdata = X_val_fold)

        # Calculate Error Rate
        fold_errors[k] <- mean(preds != y_val_fold)
      }
      grid_errors[i] <- mean(fold_errors)
    }

    # 4. Pick Winner
    best_idx <- which.min(grid_errors)
    best_laplace <- laplace_grid[best_idx]
  }

  # --- Final Training ---
  final_fit <- naivebayes::multinomial_naive_bayes(
    x = X_train_sparse,
    y = Y,
    laplace = best_laplace
  )

  y_probs <- stats::predict(final_fit, newdata = X_test_sparse, type = "prob")

  # EDGE CASE FIX: If test_vectorized has only 1 row, R might drop dimensions.
  # We force it back to a matrix to satisfy the pipeline contract.
  if (is.vector(y_probs)) {
    y_probs <- t(as.matrix(y_probs))
  }
  colnames(y_probs) <- levels(Y)
  y_pred <- stats::predict(final_fit, newdata = X_test_sparse) # Returns class labels by default

  # Ensure consistency
  y_pred_factor <- factor(y_pred, levels = levels(Y))

  return(list(
    pred = y_pred_factor,
    probs = y_probs,
    model = final_fit,
    best_lambda = NULL # Placeholder for pipeline consistency (NB doesn't use lambda)
  ))
}

#' Evaluate Model Performance (ROC and Precision-Recall)
#' @param predicted_probs Numeric vector of predicted probabilities for the positive class.
#' @param actual_classes Factor or character vector of the actual true labels.
#' @param positive_label Character string. The target class you want to evaluate.
#'
#' @return An object of class `quickSentiment_eval`, which is a list containing the following metrics:
#'   \item{target_class}{Character. The specific positive label used for the evaluation.}
#'   \item{auc_roc}{Numeric. The Area Under the Receiver Operating Characteristic curve.}
#'   \item{best_threshold_roc}{Numeric. The optimal probability threshold that maximizes Youden's J statistic.}
#'   \item{auc_pr}{Numeric. The Area Under the Precision-Recall curve.}
#'   \item{best_threshold_pr}{Numeric. The probability threshold that maximizes the F1-Score.}
#'   \item{accuracy_at_best}{Numeric. The overall accuracy of the model if `best_threshold_pr` is applied.}
#'   \item{roc}{S3 object containing the ROC curve data and metrics.}
#'   \item{prc}{S3 object containing the Precision-Recall curve data and metrics.}
#'   \item{threshold_summary}{A data frame summarizing Accuracy, Precision, Recall, and F1 at 0.1 threshold increments.}
#' @export
evaluate_performance <- function(predicted_probs, actual_classes, positive_label) {

  # 1. Vectorized Sorting
  ord <- order(predicted_probs, decreasing = TRUE)
  sorted_probs <- predicted_probs[ord]
  is_positive <- (actual_classes[ord] == positive_label)

  # 2. Cumulative sums
  tp_cum <- cumsum(is_positive)
  fp_cum <- cumsum(!is_positive)

  total_pos <- sum(is_positive)
  total_neg <- sum(!is_positive)

  if (total_pos == 0 || total_neg == 0) {
    stop("Requires atleast two class")
  }

  # --- ROC Math ---
  tpr <- tp_cum / total_pos
  fpr <- fp_cum / total_neg
  specificity <- 1 - fpr
  youden_j <- tpr + specificity - 1

  dp_roc <- diff(fpr)
  mid_tpr <- (tpr[-1] + tpr[-length(tpr)]) / 2
  auc_roc <- sum(dp_roc * mid_tpr)

  # --- Precision-Recall Math ---
  precision <- tp_cum / (tp_cum + fp_cum)
  precision[is.nan(precision)] <- 0

  f1_score <- 2 * (precision * tpr) / (precision + tpr)
  f1_score[is.nan(f1_score)] <- 0

  dp_pr <- diff(tpr)
  mid_prec <- (precision[-1] + precision[-length(precision)]) / 2
  auc_pr <- sum(dp_pr * mid_prec)

  # --- CALCULATE THRESHOLDS FIRST ---
  best_idx_pr <- which.max(f1_score)
  best_threshold_pr <- sorted_probs[best_idx_pr]

  best_idx_roc <- which.max(youden_j)
  best_threshold_roc <- sorted_probs[best_idx_roc]

  # --- Accuracy at Best PR Threshold ---
  tn_at_best <- total_neg - fp_cum[best_idx_pr]
  accuracy_at_best <- (tp_cum[best_idx_pr] + tn_at_best) / (total_pos + total_neg)

  # --- Create Data Frame ---
  curve_data <- data.frame(
    Threshold = sorted_probs,
    TPR_Recall = tpr,
    FPR = fpr,
    Precision = precision,
    F1 = f1_score
  )

  # --- NEW: DOWNSAMPLING ---
  # If data is large, skip every 10 rows but guarantee best indices are kept
  n_rows <- nrow(curve_data)
  if (n_rows > 1000) {
    keep_idx <- unique(c(1, seq(1, n_rows, by = 10), best_idx_pr, best_idx_roc, n_rows))
    keep_idx <- sort(keep_idx) # Sort to keep the curve chronological
    curve_data <- curve_data[keep_idx, ]
  }

  # --- Create dedicated S3 objects for the individual curves ---
  roc_object <- list(
    curve = curve_data,
    auc = auc_roc,
    best_threshold = best_threshold_roc
  )
  class(roc_object) <- "quickSentiment_roc"

  prc_object <- list(
    curve = curve_data,
    auc = auc_pr,
    best_threshold = best_threshold_pr
  )
  class(prc_object) <- "quickSentiment_prc"

  # --- NEW: User's Requested Threshold Summary Table ---
  # Calculate Accuracy, Precision, Recall, and F1 at exact 0.1 increments
  summary_thresholds <- seq(0, 1, by = 0.1)
  summary_table <- data.frame(
    Threshold = summary_thresholds,
    Accuracy = NA, Precision = NA, Recall = NA, F1 = NA
  )

  unsorted_is_positive <- (actual_classes == positive_label)

  for(i in 1:nrow(summary_table)) {
    th <- summary_table$Threshold[i]
    pred_pos <- predicted_probs >= th # Unsorted

    # Compare Unsorted to Unsorted
    tp <- sum(pred_pos & unsorted_is_positive)
    fp <- sum(pred_pos & !unsorted_is_positive)
    tn <- sum(!pred_pos & !unsorted_is_positive)
    fn <- sum(!pred_pos & unsorted_is_positive)

    acc <- (tp + tn) / (total_pos + total_neg)
    prec <- ifelse((tp + fp) == 0, 0, tp / (tp + fp))
    rec <- ifelse((tp + fn) == 0, 0, tp / (tp + fn))
    f1 <- ifelse((prec + rec) == 0, 0, 2 * (prec * rec) / (prec + rec))

    summary_table$Accuracy[i] <- acc
    summary_table$Precision[i] <- prec
    summary_table$Recall[i] <- rec
    summary_table$F1[i] <- f1
  }

  # --- Package the Results (Flattened correctly!) ---
  results <- list(
    target_class       = positive_label,
    auc_roc            = auc_roc,
    best_threshold_roc = best_threshold_roc,
    auc_pr             = auc_pr,
    best_threshold_pr  = best_threshold_pr,
    accuracy_at_best   = accuracy_at_best,
    roc                = roc_object, # The ROC S3 Object
    prc                = prc_object,  # The PRC S3 Object
    threshold_summary  = summary_table
  )

  structure(results, class = "quickSentiment_eval")
}

#' Predict Sentiment on New Data Using a Saved Pipeline Artifact
#'
#' This is a generic prediction function that handles different model types
#' and ensures consistent preprocessing and vectorization for new, unseen text.
#'
#' @param pipeline_object A list object returned by the main `pipeline()` function.
#'   It must contain the trained model, DFM template, preprocessing function,
#'   and n-gram settings.
#' @param text_column A string specifying the column name of the text to predict.
#' @param threshold Numeric. Optional custom threshold for binary classification.
#'   If NULL, uses the optimized threshold from training (if available).
#'
#' @return A data frame containing the `predicted_class` and probability columns.
#'
#' @importFrom stats predict
#' @importFrom xgboost xgb.DMatrix
#'
#' @export
#' @examples
#' \donttest{
#' if (exists("my_artifacts")) {
#'   dummy_df <- data.frame(text = c("loved it", "hated it"), stringsAsFactors = FALSE)
#'   preds <- predict_sentiment(my_artifacts, df = dummy_df, text_column = "text")
#'  }
#' }
#'
predict_sentiment <- function(pipeline_object,
                       text_column,
                       threshold = 0.5) {

  # --- 1. Extract the necessary artifacts from the pipeline object ---
  final_model <- pipeline_object$trained_model
  dfm_template <- pipeline_object$dfm_template
  class_levels <- pipeline_object$class_levels #for XGBoost since it does only numeric for Y

  # --- 2. Preprocess and vectorize the new data ---
  message("--- Preparing new data for prediction ---\n")

  new_dfm <- BOW_test(
    text_column,
    dfm_template)

  # --- 3. Multi-Class Guardrail ---
  # If there are more than 2 classes, thresholds mathematically do not apply.
  if (length(class_levels) > 2) {
    if (threshold != 0.5) {
      warning("Threshold argument is ignored for multi-class prediction. Assigning class based on highest probability.")
    }
    threshold <- NULL # Pass NULL so route_prediction knows to just use max probability
  }

  # --- 4. Conditional Prediction based on Model Class ---
  message("--- Making Predictions ---\n")

  results <- route_prediction(
    new_data_vectorized = new_dfm,
    qs_artifact = pipeline_object,
    model_type = pipeline_object$model_type,
    Y_levels = class_levels,
    threshold = threshold
  )
  #--- 5.  Output
  final_output <- data.frame(predicted_class = results$pred)

  prob_df <- as.data.frame(results$probs)
  colnames(prob_df) <- paste0("prob_", colnames(prob_df))

  # Bind the probabilities to the class predictions
  final_output <- cbind(final_output, prob_df)

  message("--- Prediction Complete ---\n")
  return(final_output)

}

#' Train a Gradient Boosting Model using XGBoost
#'
#' This function trains a model using the xgboost package. It is highly
#' efficient and natively supports sparse matrices, making it ideal for text data.
#' It automatically handles both binary and multi-class classification problems.
#'
#' @param train_vectorized The training feature matrix (e.g., a `dfm` from quanteda).
#' @param Y The response variable for the training set. Should be a factor.
#' @param test_vectorized The test feature matrix, which must have the same
#'   features as `train_vectorized`.
#' @param parallel Logical
#' @param tune Logical
#'
#' @return A list containing four elements:
#'   \item{pred}{A vector of class predictions for the test set.}
#'   \item{probs}{A matrix of predicted probabilities.}
#'   \item{model}{The final, trained `xgb.Booster` model object.}
#'   \item{best_lambda}{Placeholder (NULL) for pipeline consistency.}
#'
#' @importFrom xgboost xgb.train xgb.DMatrix xgb.cv
#' @importFrom stats predict
#' @importFrom methods as
#'
#' @export
#' @examples
#' \dontrun{
#' # Create dummy vectorized training and test data
#' train_matrix <- matrix(runif(100), nrow = 10, ncol = 10)
#' test_matrix <- matrix(runif(50), nrow = 5, ncol = 10)
#'
#' # Provide column names (vocabulary) required by xgboost
#' colnames(train_matrix) <- paste0("word", 1:10)
#' colnames(test_matrix) <- paste0("word", 1:10)
#'
#' y_train <- factor(sample(c("P", "N"), 10, replace = TRUE))
#'
#' # Run xgboost model
#' model_results <- xgb_model(train_matrix, y_train, test_matrix)
#' }

xgb_model <- function(train_vectorized, Y, test_vectorized, parallel = FALSE,tune = FALSE) {

  message("\n--- Training XGBoost Model ---\n")

  # --- CHANGE 1: Prepare the Y variable ---
  # Get the number of classes from the factor levels
  num_classes <- length(levels(Y))
  # Convert factor Y ('neg', 'neu', 'pos') to numeric ('1','2','3') and then to zero-indexed ('0','1','2')
  y_train_numeric <- as.numeric(Y) - 1

  # Setup parallel threads
  threads <- if (isTRUE(parallel)) parallel::detectCores() - 1 else 1

  # --- MEMORY SAFETY: Cast quanteda DFM to dgCMatrix ---
  X_train_sparse <- methods::as(train_vectorized, "dgCMatrix")
  X_test_sparse <- methods::as(test_vectorized, "dgCMatrix")

  # This switch ensures Binary models allow Thresholds, while Multi-class works normally.
  if (num_classes == 2) {
    params <- list(
      objective = "binary:logistic",
      eval_metric = "logloss",
      eta = 0.1,
      max_depth = 4,
      nthread = threads
    )
  } else {
    params <- list(
      objective = "multi:softprob",
      eval_metric = "mlogloss",
      num_class = num_classes,
      eta = 0.1,
      max_depth = 4,
      nthread = threads
    )
  }

  # Prepare DMatrix objects (this part is the same)
  dtrain <- xgboost::xgb.DMatrix(data = X_train_sparse, label = y_train_numeric)
  dtest <- xgboost::xgb.DMatrix(data = X_test_sparse)

  # --- CONDITIONAL TUNING LOGIC ---
  if (isTRUE(tune)) {
    message("  - Using xgb.cv to find optimal nrounds (early stopping)...")

    cv_results <- xgboost::xgb.cv(
      params = params,
      data = dtrain,
      nfold = 5,
      nrounds = 500,
      early_stopping_rounds = 20,
      verbose = 0
    )

    best_nrounds <- cv_results$best_iteration

  } else {
    # Fast route: Use your original default
    best_nrounds <- 100
  }

  # Train the model (this part is the same)
  xgb_model <- xgboost::xgb.train(
    params = params,
    data = dtrain,
    nrounds = best_nrounds,
    verbose = 0
  )

  # --- CHANGE 3: Process the predictions ---
  # predict() now returns a matrix of probabilities (rows=samples, cols=classes)

  y_pred_prob_matrix <- predict(xgb_model, newdata = dtest, reshape = TRUE)

  if (num_classes == 2) {
    # Binary: Output is vector of probs for Class 1
    # Use default 0.5 threshold for the internal evaluation report
    prob_v <- as.vector(y_pred_prob_matrix)
    y_probs <- matrix(c(1 - prob_v, prob_v), ncol = 2)
    colnames(y_probs) <- levels(Y)

    y_pred_numeric <- ifelse(prob_v > 0.5, 1, 0)

    } else {
    # Multi-class: Output is Matrix
      y_probs <- y_pred_prob_matrix
      colnames(y_probs) <- levels(Y)
      y_pred_numeric <- max.col(y_probs) - 1
  }


  # Map the numeric predictions back to the original factor labels for evaluation
  y_pred_factor <- factor(y_pred_numeric, levels = 0:(num_classes-1), labels = levels(Y))
 
  # --- ENFORCE THE CONTRACT (this part is the same) ---
  results <- list(
    pred = y_pred_factor,
    probs = y_probs,
    model = xgb_model,
    best_lambda = NULL
  )

  return(results)
}

#' Standard Negation Words for Sentiment Analysis
#'
#' A character vector of 25 common negation words. These words are automatically
#' protected by the \code{\link{pre_process}} function when \code{retain_negations = TRUE}
#' to prevent standard stopword lists from destroying sentiment polarity.
#'
#' @export
qs_negations <- c("no", "not", "nor", "neither", "never", "none", "cannot",
                  "aren't", "couldn't", "didn't", "doesn't", "don't",
                  "hadn't", "hasn't", "haven't", "isn't", "mightn't",
                  "mustn't", "needn't", "shan't", "shouldn't", "wasn't",
                  "weren't", "won't", "wouldn't")

#' Preprocess a Vector of Text Documents
#'
#' This function provides a comprehensive and configurable pipeline for cleaning
#' raw text data. It handles a variety of common preprocessing steps including
#' removing URLs and HTML, lowercasing, stopword removal, and lemmatization.
#'
#' @param doc_vector A character vector where each element is a document.
#' @param remove_brackets A logical value indicating whether to remove text in square brackets.
#' @param remove_urls A logical value indicating whether to remove URLs and email addresses.
#' @param remove_html A logical value indicating whether to remove HTML tags.
#' @param remove_nums A logical value indicating whether to remove numbers.
#' @param remove_emojis_flag A logical value indicating whether to remove common emojis.
#' @param to_lowercase A logical value indicating whether to convert text to lowercase.
#' @param remove_punct A logical value indicating whether to remove punctuation.
#' @param remove_stop_words A logical value indicating whether to remove English stopwords.
#' @param custom_stop_words A character vector of additional custom words to remove (e.g., c("rt", "via")). Default is NULL.
#' @param keep_words A character vector of words to protect from deletion (e.g., c("no", "not", "nor")). Default is NULL.
#' @param lemmatize A logical value indicating whether to lemmatize words to their dictionary form.
#' @param retain_negations Logical. If \code{TRUE} (the default), automatically protects common negation words (e.g., "not", "no", "never") from being deleted by the standard stopword list to preserve sentiment context.
#'
#' @return A character vector of the cleaned and preprocessed text.
#'
#' @importFrom stringr str_replace_all str_remove_all str_squish
#' @importFrom quanteda tokens tokens_select
#' @importFrom textstem lemmatize_strings
#' @importFrom stopwords stopwords
#'
#'
#' @export
#' @examples
#' raw_text <- c(
#'   "This is a <b>test</b>! Visit https://example.com",
#'   "Email me at test.user@example.org [important]"
#' )
#'
#' # Basic preprocessing with defaults
#' clean_text <- pre_process(raw_text)
#' print(clean_text)
#'
#' # Keep punctuation and stopwords
#' clean_text_no_stop <- pre_process(
#'   raw_text,
#'   remove_stop_words = FALSE,
#'   remove_punct = FALSE
#' )
#' print(clean_text_no_stop)
pre_process <- function(doc_vector,
  remove_brackets = TRUE,
  remove_urls = TRUE,
  remove_html = TRUE,
  remove_nums = FALSE,
  remove_emojis_flag = TRUE,
  to_lowercase = TRUE,
  remove_punct = TRUE,
  remove_stop_words = TRUE,
  custom_stop_words = NULL,
  keep_words = NULL,
  lemmatize = TRUE,
  retain_negations = TRUE) {

  # Validate input
  if (!is.atomic(doc_vector)) {
    stop("`doc_vector` must be an atomic vector (usually character).", call. = FALSE)
  }

  # Preserve NA positions and coerce safely
  doc_vector <- as.character(doc_vector)


  # --- Stage 1: String-level cleaning ---

  if (remove_brackets) {
    doc_vector <- stringr::str_replace_all(doc_vector, "\\[[^\\]]*\\]", " ")
  }
  if (remove_urls) {
    doc_vector <- stringr::str_replace_all(doc_vector, "(?i)\\bhttps?://\\S+\\b|\\bwww\\.\\S+\\b", "")
  }
  if (remove_html) {
    doc_vector <- stringr::str_replace_all(doc_vector, "<[^>]+>", "")
  }
  if (remove_emojis_flag) {
    doc_vector <- stringr::str_remove_all(doc_vector, "[\\U0001F600-\\U0001F64F]|[\\U0001F300-\\U0001F5FF]|[\\U0001F680-\\U0001F6FF]|[\\U0001F1E0-\\U0001F1FF]")
  }
  doc_vector <- stringr::str_squish(doc_vector)

  # --- Stage 2: Tokenization and token-level operations ---
  # A more efficient, standard workflow is to perform all string operations first.
  if (to_lowercase) {
    doc_vector <- base::tolower(doc_vector)
  }


  # Now, tokenize and remove stopwords/punctuation/numbers
  toks <- quanteda::tokens(doc_vector,
                          remove_punct = remove_punct,
                          remove_numbers = remove_nums)

  # --- Stopword Removal Logic ---
  if (remove_stop_words || !is.null(custom_stop_words)) {
    stops_to_remove <- if (remove_stop_words) quanteda::stopwords("en") else character(0)

    # 1. Handle Negations Explicitly
    if (retain_negations && remove_stop_words) {
      message("quickSentiment: Retaining negation words (e.g., 'not', 'no', 'never') to preserve sentiment polarity. To apply the strict stopword list instead, set `retain_negations = FALSE`. View qs_negations for more")


      # Protect negations and user's keep_words
      words_to_protect <- unique(c(tolower(keep_words), qs_negations))
      stops_to_remove <- setdiff(stops_to_remove, words_to_protect)

    } else {
      # If they turned it off, just protect their keep_words
      stops_to_remove <- setdiff(stops_to_remove, tolower(keep_words))
    }

    # 2. Add custom_stop_words LAST (The User's Absolute Override)
    if (!is.null(custom_stop_words)) {
      stops_to_remove <- unique(c(stops_to_remove, tolower(custom_stop_words)))
    }

    # 3. Filter the tokens
    toks <- quanteda::tokens_select(toks, pattern = stops_to_remove, selection = "remove")
  }


  # --- Final step: Convert tokens back to a single string ---
  out <- vapply(toks, function(x) paste(x, collapse = " "), character(1))
  if (lemmatize) {
    out <- textstem::lemmatize_strings(out)
  }
  out <- stringr::str_squish(out)
  return(out)


  }

#' functions/random_forest_fast.R
#' Train a Random Forest Model using Ranger
#'
#' This function trains a Random Forest model using the high-performance ranger
#' package. It natively utilizes sparse matrices (dgCMatrix) to avoid memory
#' exhaustion and utilizes Out-Of-Bag (OOB) error for rapid hyperparameter tuning.
#' @param train_vectorized The training feature matrix (e.g., a `dfm` from quanteda).
#' @param Y The response variable for the training set. Should be a factor.
#' @param test_vectorized The test feature matrix, which must have the same
#'   features as `train_vectorized`.
#' @param parallel Logical
#' @param tune Logical. If TRUE, tunes `mtry` using native OOB error
#'
#' @return A list containing four elements:
#'   \item{pred}{A vector of class predictions for the test set.}
#'   \item{probs}{A matrix of predicted probabilities.}
#'   \item{model}{The final, trained `ranger` model object.}
#'   \item{best_lambda}{Placeholder (NULL) for pipeline consistency.}
#'
#' @importFrom ranger ranger
#' @importFrom stats predict
#' @importFrom methods as
#'
#' @export
#' @examples
#' \dontrun{
#' # Create dummy vectorized training and test data
#' train_matrix <- matrix(runif(100), nrow = 10, ncol = 10)
#' test_matrix <- matrix(runif(50), nrow = 5, ncol = 10)
#'
#' # Provide column names (vocabulary) required by ranger
#' colnames(train_matrix) <- paste0("word", 1:10)
#' colnames(test_matrix) <- paste0("word", 1:10)
#'
#' y_train <- factor(sample(c("P", "N"), 10, replace = TRUE))
#'
#' # Run random forest model
#' model_results <- rf_model(train_matrix, y_train, test_matrix)
#' }
rf_model <- function(train_vectorized, Y, test_vectorized, parallel = FALSE, tune = FALSE) {

  message("\n--- Training Random Forest Model (ranger) ---\n")

  p <- ncol(train_vectorized)
  threads <- if (isTRUE(parallel)) parallel::detectCores() - 1 else 1

  # Check if anyone passed a dataframe
  if (is.data.frame(train_vectorized)) {
    train_vectorized <- as.matrix(train_vectorized)
    test_vectorized <- as.matrix(test_vectorized)
  }
  # Cast the quanteda DFM directly to a standard dgCMatrix
  X_train_matrix<- methods::as(train_vectorized, "dgCMatrix")
  X_test_matrix <- methods::as(test_vectorized, "dgCMatrix")

  if (isTRUE(tune)) {
    message("  - Tuning mtry using native Out-Of-Bag (OOB) error(this may take time)...")

    # Define a grid for mtry.
    # We test: sqrt(p), 2*sqrt(p), and p/3
    # Define a smart grid for mtry based on feature count (p)
    mtry_grid <- unique(floor(c(sqrt(p), sqrt(p) * 2, p / 3)))
    best_oob_error <- Inf
    ranger_model <- NULL

    for (m in mtry_grid) {
      # Train a model directly on the sparse matrices using x/y
      temp_model <- ranger::ranger(
        x = X_train_matrix,
        y = Y,
        num.trees = 200,
        mtry = m,
        num.threads = threads,
        importance = "impurity",
        probability = TRUE,   # We need probabilities for ROC/AUC later
        verbose = FALSE
      )
     # ranger saves the OOB error automatically!
      # (Since probability=TRUE, this is the Brier score)
      current_error <- temp_model$prediction.error
      # Keep the model if it has the lowest error so far
      if (current_error < best_oob_error) {
        best_oob_error <- current_error
        ranger_model <- temp_model
      }
    }

  } else {
    # Fast Heuristic Route

    actual_mtry <- min(p, floor(sqrt(p) * 2))

    ranger_model <- ranger::ranger(
      x = X_train_matrix,
      y = Y,
      num.trees = 200,
      mtry = actual_mtry,
      num.threads = threads,
      importance = "impurity",
      verbose = FALSE,
      probability = TRUE
    )
  }

  #Required for TF-IDF
  expected_cols <- colnames(X_train_matrix)

  # Handle total name wipe
  if (is.null(colnames(X_test_matrix)) && ncol(X_test_matrix) == length(expected_cols)) {
    colnames(X_test_matrix) <- expected_cols
  }

  current_cols <- colnames(X_test_matrix)
  if (!is.null(current_cols)) {
    missing_cols <- setdiff(expected_cols, current_cols)
    if (length(missing_cols) > 0) {
      # Build a sparse matrix of 0s for the missing words
      zero_padding <- Matrix::Matrix(0,
                                     nrow = nrow(X_test_matrix),
                                     ncol = length(missing_cols),
                                     sparse = TRUE)
      # Safely assign names to the padding
      dimnames(zero_padding) <- list(rownames(X_test_matrix), missing_cols)

      # Bind the missing columns back
      X_test_matrix <- cbind(X_test_matrix, zero_padding)

      # THE CRITICAL FIX: Force the names back on!
      # (Prevents cbind from stripping names)
      colnames(X_test_matrix) <- c(current_cols, missing_cols)
    }
  } else {
    stop("Fatal error: Test matrix lost column names and dimensions do not match training data.")
  }

   X_test_matrix <- X_test_matrix[, expected_cols, drop = FALSE]


  predictions_obj <- predict(ranger_model, data = X_test_matrix)
  probs_matrix <- predictions_obj$predictions

  # CONVERT TO CLASSES (For Pipeline Internal Report) ---
  if (ncol(probs_matrix) == 2) {
    # Binary: If Prob(Class2) > 0.5, predict Class2
    y_pred_idx <- ifelse(probs_matrix[, 2] > 0.5, 2, 1)
  } else {
    # Multiclass: Pick column with max probability
    y_pred_idx <- max.col(probs_matrix)
  }

  # Convert indices back to original Factor Levels
  y_pred_factor <- factor(levels(Y)[y_pred_idx], levels = levels(Y))
  message("--- Random Forest complete. Returning results. ---\n")
  return(list(
    pred = y_pred_factor,
    probs = probs_matrix,
    model = ranger_model,
    best_lambda = NULL
  ))
}

#' Internal Router for Model Inference
#'
#' This is an internal helper function that routes a preprocessed sparse matrix
#' to the correct machine learning model. It guarantees that regardless of which
#' algorithm was trained, the output predictions and probability matrices follow
#' a strict, standardized format.
#'
#' @param new_data_vectorized A quanteda `dfm` containing the new text data.
#' @param qs_artifact The trained model list returned by the quickSentiment pipeline.
#' @param model_type A character string: "logit", "rf", "xgb", or "nb".
#' @param Y_levels A character vector of the original target variable levels.
#' @param threshold Numeric. The probability threshold for the positive class.
#'
#' @return A list containing:
#'   \item{pred}{A factor vector of class predictions.}
#'   \item{probs}{A properly named matrix of predicted probabilities.}
#'
#' @importFrom stats predict
#' @importFrom methods as
#' @importFrom xgboost xgb.DMatrix
#'
#' @noRd
route_prediction <- function(new_data_vectorized, qs_artifact, model_type, Y_levels, threshold) {

  # --- 1. GLOBAL MEMORY SAFETY ---
  # Force the DFM to a pure mathematical sparse matrix for all models
  X_new_sparse <- methods::as(new_data_vectorized, "dgCMatrix")
  num_classes <- length(Y_levels)

  # --- 2. MODEL ROUTING EXECUTION ---

  if (model_type == "rf") {

    expected_cols <- qs_artifact$trained_model$forest$independent.variable.names

    # 2. Handle total name wipe
    if (is.null(colnames(X_new_sparse)) && ncol(X_new_sparse) == length(expected_cols)) {
      colnames(X_new_sparse) <- expected_cols
    }

    current_cols <- colnames(X_new_sparse)
    if (!is.null(current_cols)) {
      missing_cols <- setdiff(expected_cols, current_cols)
      if (length(missing_cols) > 0) {
        # Build a sparse matrix of 0s for the missing words
        zero_padding <- Matrix::Matrix(0,
                                       nrow = nrow(X_new_sparse),
                                       ncol = length(missing_cols),
                                       sparse = TRUE)
        # Safely assign names to the padding
        dimnames(zero_padding) <- list(rownames(X_new_sparse), missing_cols)

        # Bind the missing columns back
        X_new_sparse <- cbind(X_new_sparse, zero_padding)

        # THE CRITICAL FIX: Force the names back on!
        colnames(X_new_sparse) <- c(current_cols, missing_cols)
      }
    } else {
      stop("Fatal error: New data matrix lost column names.")
    }

    # 3. Perfectly order the columns safely
    X_new_sparse <- X_new_sparse[, expected_cols, drop = FALSE]
    # ------------------------------------

    # Ranger returns a specific predictions object
    predictions_obj <- stats::predict(qs_artifact$trained_model, data = X_new_sparse)
    probs_matrix <- predictions_obj$predictions

  } else if (model_type == "xgb") {

    # XGBoost requires its own custom DMatrix format
    dtest <- xgboost::xgb.DMatrix(data = X_new_sparse)
    raw_probs <- stats::predict(qs_artifact$trained_model, newdata = dtest, reshape = TRUE)

    # XGB binary returns a single vector of positive probabilities
    if (num_classes == 2) {
      prob_v <- as.vector(raw_probs)
      probs_matrix <- matrix(c(1 - prob_v, prob_v), ncol = 2)
    } else {
      probs_matrix <- raw_probs
    }

  } else if (model_type == "logit") {

    # glmnet requires the saved best_lambda penalty parameter
    raw_probs <- stats::predict(qs_artifact$trained_model, newx = X_new_sparse,
                                type = "response", s = qs_artifact$best_lambda)

    # glmnet binary returns a single column of positive probabilities
    if (num_classes == 2) {
      prob_v <- as.vector(raw_probs)
      probs_matrix <- matrix(c(1 - prob_v, prob_v), ncol = 2)
    } else {
      # glmnet multiclass returns a 3D array (n_rows x n_classes x 1). Extract the 1st slice.
      probs_matrix <- raw_probs[, , 1]
    }

  } else if (model_type == "nb") {

    # Naive Bayes naturally returns a clean probability matrix
    probs_matrix <- stats::predict(qs_artifact$trained_model, newdata = X_new_sparse, type = "prob")

  } else {
    stop("Invalid model_type. Must be 'rf', 'xgb', 'logit', or 'nb'.")
  }

  # --- 3. EDGE CASE SAFETY NET ---
  # If the user only predicted ONE sentence, R drops dimensions to a vector.
  # We force it back into a 1-row, 2-column (or n-column) matrix.
  if (is.vector(probs_matrix) || length(dim(probs_matrix)) == 1) {
    probs_matrix <- t(as.matrix(probs_matrix))
  }

  # We apply the global standard to the column names so the wrapper can easily bind them
  colnames(probs_matrix) <- Y_levels

  # --- 4. CONVERT PROBABILITIES TO CLASSES ---
  if (is.null(threshold)) {
    threshold <- 0.5
  }

  if (ncol(probs_matrix) == 2) {
    # Binary: If Prob(Class2) > threshold, predict Class2
    y_pred_idx <- ifelse(probs_matrix[, 2] >= threshold, 2, 1)
  } else {
    # Multiclass: Pick column with max probability (threshold is naturally ignored)
    y_pred_idx <- max.col(probs_matrix, ties.method = "first")
  }

  # Rebuild the final factor vector safely
  y_pred_factor <- factor(Y_levels[y_pred_idx], levels = Y_levels)

  return(list(
    pred = y_pred_factor,
    probs = probs_matrix
  ))
}

#' Run a Full Text Classification Pipeline on Preprocessed Text
#'
#' This function takes a data frame with pre-cleaned text and handles the
#' data splitting, vectorization, model training, and evaluation.
#'
#' @param vect_method A string specifying the vectorization method.
#'   Defaults to \code{"bag_of_words"}.
#'   \itemize{
#'     \item \code{"bag_of_words"} (Alias: \code{"bow"}) - Standard count of words.
#'     \item \code{"term_frequency"} (Alias: \code{"tf"}) - Normalized counts.
#'     \item \code{"tfidf"} (Alias: \code{"tf-idf"}) - Term Frequency-Inverse Document Frequency.
#'     \item \code{"binary"} - Presence/Absence (1/0).
#'   }
#' @param model_name A string specifying the model to train.
#'   Defaults to \code{"logistic_regression"}.
#'   \itemize{
#'     \item \code{"random_forest"} (Alias: \code{"rf"})
#'     \item \code{"xgboost"} (Alias: \code{"xgb"})
#'     \item \code{"logistic_regression"} (Alias: \code{"logit"}, \code{"glm"})
#'   }
#' @param text_vector A character vector containing the **preprocessed** text.
#' @param sentiment_vector A vector or factor containing the target labels (e.g., ratings).
#' @param n_gram The n-gram size to use for BoW/TF-IDF. Defaults to 1.
#' @param parallel If TRUE, runs model training in parallel. Default FALSE.
#' @param tune Logical. If TRUE, the pipeline will perform hyperparameter tuning
#'    for the selected model. Defaults to FALSE. [NEW]
#' @return A list containing the trained model object, the DFM template,
#'   class levels, and a comprehensive evaluation report.
#' @importFrom parallel detectCores makeCluster stopCluster
#' @importFrom doParallel registerDoParallel
#' @importFrom foreach registerDoSEQ
#' @export
#' @examples
#' df <- data.frame(
#'   text = c("good product", "excellent", "loved it", "great quality",
#'            "bad service", "terrible", "hated it", "awful experience",
#'            "not good", "very bad", "fantastic", "wonderful"),
#'   y = c("P", "P", "P", "P", "N", "N", "N", "N", "N", "N", "P", "P")
#' )
#'
#'
#' out <- pipeline("bow", "naive_bayes",  text_vector = df$text, sentiment_vector = df$y)
#'
pipeline <- function(vect_method,
  model_name,
  text_vector,
  sentiment_vector,
  n_gram = 1,
  tune = FALSE,
  parallel=FALSE) {

    stopf <- function(...) stop(sprintf(...), call. = FALSE)

    # --- 1. CLEAN & TRANSLATE ARGUMENTS ---
    #to lower and trim ensures small typos don't break the function and allows for more flexible input
    vect_method <- tolower(trimws(vect_method))
    model_name  <- tolower(trimws(model_name))
    n_gram <- as.integer(n_gram)

    # ALIASING: VECTORIZERS
    if (vect_method == "bow") vect_method <- "bag_of_words"
    if (vect_method == "tf")  vect_method <- "term_frequency"
    if (vect_method == "tf-idf")  vect_method <- "tfidf"

    # ALIASING: Convert shortcuts to official names
    if (model_name == "rf")    model_name <- "random_forest"
    if (model_name == "xgb")   model_name <- "xgboost"
    if (model_name == "nb")   model_name <- "naive_bayes"
    if (model_name %in% c("logit","glm","logistic")) model_name <- "logistic_regression"


  # Ensure Vector method is in Input
    allowed_vect <- c("bag_of_words", "binary", "term_frequency", "tfidf")
    if (!vect_method %in% allowed_vect) {
      stopf("Vectorizer '%s' is not supported. Use: %s.",
            vect_method, paste(allowed_vect, collapse = ", "))
    }

  # Ensure Model name is in Input
    allowed_models <- c("logistic_regression", "random_forest", "xgboost", "naive_bayes")
    if (!model_name %in% allowed_models) {
      stopf("Model '%s' is not supported. Use: %s.",
            model_name, paste(allowed_models, collapse = ", "))
    }

  # Model list that can be used
    models_list <- list(
      logistic_regression = logit_model,
      random_forest       = rf_model,
      xgboost             = xgb_model,
      naive_bayes         = nb_model
    )


    if (!is.atomic(text_vector)) stopf("`text_vector` must be an atomic vector (usually character).")
    if (!is.atomic(sentiment_vector) && !is.factor(sentiment_vector)) stopf("`sentiment_vector` must be a vector or factor.")

    if (length(text_vector) != length(sentiment_vector)) {
      stopf("Length mismatch: `text_vector` has %d elements, but `label_vector` has %d elements.",
            length(text_vector), length(sentiment_vector))
    }

  #FOR GLMNET ONLY
  if (isTRUE(parallel)) {
    # Dynamically detect cores and register
    n_cores <- parallel::detectCores() - 1
    cl <- parallel::makeCluster(n_cores)
    doParallel::registerDoParallel(cl)

    # Ensure the cluster stops even if the code crashes
    on.exit(parallel::stopCluster(cl), add = TRUE)
    on.exit(foreach::registerDoSEQ(), add = TRUE)
  }

  message(paste0("--- Running Pipeline: ", toupper(vect_method), " + ", toupper(model_name), " ---\n"))

  #drop unnecessary columns and ensure sentiment is a factor
  initial_length <- length(text_vector)
  # Force formatting safely
  text_vector <- trimws(as.character(text_vector))
  text_vector[text_vector == ""] <- NA_character_
  sentiment_vector <- as.factor(sentiment_vector)

  # Find valid indices (where neither text nor label is NA)
  valid_idx <- !is.na(text_vector) & !is.na(sentiment_vector)

  # Subset vectors simultaneously
  text_vector <- text_vector[valid_idx]
  sentiment_vector <- sentiment_vector[valid_idx]

  rows_dropped <- initial_length - length(text_vector)

  if (rows_dropped > 0) {
    warning(sprintf("Dropped %d element(s) with missing/empty text or missing sentiments(labels).", rows_dropped),
            call. = FALSE)
  }
  if (length(text_vector) < 5) stopf("Not enough valid data after filtering (%d).", length(text_vector))
  if (nlevels(sentiment_vector) < 2) stopf("Need at least 2 sentiment classes after filtering.")


  # --- 2. TRAIN/TEST SPLIT ---

  # Base R Stratified Split
  train_idx <- unlist(lapply(split(seq_along(sentiment_vector), sentiment_vector), function(idx) {
    sample(idx, size = round(0.8 * length(idx)))
  }))

  # Ensure it's a clean numeric vector
  train_idx <- as.numeric(train_idx)

  text_train <- text_vector[train_idx]
  text_test <- text_vector[-train_idx]

  # Split Labels
  y_train <- sentiment_vector[train_idx]
  y_test  <- sentiment_vector[-train_idx]

  message(paste0("Data split: ", length(text_train), " training elements, ", length(text_test), " test elements.\n"))

  # --- 3. VECTORIZATION ---
  # This now operates on the pre-cleaned text column
  message(sprintf("Vectorizing with %s (ngram=%d)...", toupper(vect_method), n_gram))

  fit <- BOW_train(text_train,
                   weighting_scheme = vect_method,
                   ngram_size = n_gram)
  X_train <- fit$dfm_template

  X_test <- BOW_test(text_test,fit)



  # --- 4. MODEL TRAINING & PREDICTION ---

  model_results <- models_list[[model_name]](
    X_train,
    y_train,
    X_test,
    parallel = parallel,
    tune=tune
    )
  if (is.null(model_results$model) || is.null(model_results$pred)) {
    stopf("Model function '%s' must return a list with elements `model` and `pred`.", model_name)
  }


  # --- 5. PACKAGE ARTIFACTS ---

 internal_type_map <- c(
  "logistic_regression" = "logit",
  "random_forest" = "rf",
  "xgboost" = "xgb",
  "naive_bayes" = "nb"
)
  # ---  RETURN FINAL RESULTS ---
  final_output <- list(
      trained_model = model_results$model,
      model_type    = internal_type_map[model_name],
      best_lambda   = model_results$best_lambda, # <- for glmnet only
      dfm_template = fit,
      class_levels = levels(y_test),
      ngram_size_used = n_gram,
      probs = model_results$probs,
      y_test= y_test
      )

  # --- 6. USER GUIDANCE MESSAGE ---
  baseline_accuracy <- mean(model_results$pred== y_test)
  acc_pct <- round(baseline_accuracy * 100, 2)

  message("\n======================================================")
  message(" --- quickSentiment Pipeline Complete ---")
  message(sprintf(" Model Type: %s", toupper(model_name)))
  message(sprintf(" Vectorizer: %s (ngram=%d)", toupper(vect_method), n_gram))
  message(sprintf(" Test Set Size: %d rows", length(y_test)))
  message(sprintf(" Accuracy of %s%% under baseline threshold.", acc_pct))
  message("======================================================\n")

  return(final_output)
}

---
title: "Introduction to quickSentiment"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to quickSentiment}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# --- 1. SETUP: LOAD LIBRARIES ---
# -------------------------------------------------------------------
```{r, include = FALSE}
library(quickSentiment)

```

```{r setup}
library(doParallel)
# CRAN limits the number of cores used during package checks
cores <- min(2, parallel::detectCores())
registerDoParallel(cores = cores)
```

# --- 2. LOAD AND PREPARE TRAINING DATA ---
```{r}
# Look for the file in the installed package first
csv_path <- system.file("extdata", "tweets.csv", package = "quickSentiment")

# Fallback for when you are building the package locally
if (csv_path == "") {
  csv_path <- "../inst/extdata/tweets.csv"
}
tweets <- read.csv(csv_path)
set.seed(123)
```

# --- 3. PREPROCESS THE TEXT ---
# -------------------------------------------------------------------
## Use the pre_process() function from our package to clean the raw text.
## This step is done externally to the main pipeline, allowing you to reuse
## the same cleaned text for multiple different models or analyses in the future.


```{r}
tweets$cleaned_text <- pre_process(tweets$Tweet)
tweets$sentiment = ifelse(tweets$Avg>0,'P','N')
```

# --- 4. RUN THE MAIN TRAINING PIPELINE ---
# -------------------------------------------------------------------
## This is the core of the package. We call the main pipeline() function
## to handle the train/test split, vectorization, model training, and evaluation.

```{r}
result <- pipeline(
  # --- Define the vectorization method ---
  # Options: "bow" (raw counts), "tf" (term frequency), "tfidf", "binary"
  vect_method = "tf",
  
  # --- Define the model to train ---
  # Options: "logit", "rf", "xgb","nb"
  model_name = "rf",
  
  # --- Specify the data and column names ---
  text_vector = tweets$cleaned_text  ,   # The column with our preprocessed text
  sentiment_vector = tweets$sentiment,    # The column with the target variable
  
  # --- Set vectorization options ---
  # Use n_gram = 2 for unigrams + bigrams, or 1 for just unigrams
  n_gram = 1,
  parallel = cores
)
```
# ===================================================================
# --- 5. EVALUATE THE RESULTS
# ===================================================================
## Get the AUC, ROC and Accuracy at Decile Threshold

```{r}
evaluate_result<- evaluate_performance(result$probs[,2],result$y_test,"P")
evaluate_result
```


```{r}
plot(evaluate_result$roc)
```


```{r}
plot(evaluate_result$prc)
```
# ===================================================================
# --- 6. PREDICTION ON NEW, UNSEEN DATA ---
# ===================================================================
## The training is complete. The 'result' object now contains our trained
## model and all the necessary "artifacts" for prediction.
```{r}

predicted_tweets <- predict_sentiment(
  pipeline_object = result,
  tweets$cleaned_text
)
head(predicted_tweets)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prediction.R
\name{predict_sentiment}
\alias{predict_sentiment}
\title{Predict Sentiment on New Data Using a Saved Pipeline Artifact}
\usage{
predict_sentiment(pipeline_object, text_column, threshold = 0.5)
}
\arguments{
\item{pipeline_object}{A list object returned by the main `pipeline()` function.
It must contain the trained model, DFM template, preprocessing function,
and n-gram settings.}

\item{text_column}{A string specifying the column name of the text to predict.}

\item{threshold}{Numeric. Optional custom threshold for binary classification.
If NULL, uses the optimized threshold from training (if available).}
}
\value{
A data frame containing the `predicted_class` and probability columns.
}
\description{
This is a generic prediction function that handles different model types
and ensures consistent preprocessing and vectorization for new, unseen text.
}
\examples{
\donttest{
if (exists("my_artifacts")) {
  dummy_df <- data.frame(text = c("loved it", "hated it"), stringsAsFactors = FALSE)
  preds <- predict_sentiment(my_artifacts, df = dummy_df, text_column = "text")
 }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.quickSentiment_eval.R
\name{plot.quickSentiment_roc}
\alias{plot.quickSentiment_roc}
\title{Plot ROC Curve}
\usage{
\method{plot}{quickSentiment_roc}(x, ...)
}
\arguments{
\item{x}{An object of class `quickSentiment_roc`.}

\item{...}{Additional graphical parameters.}
}
\description{
Plot ROC Curve
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.quickSentiment_eval.R
\name{print.quickSentiment_eval}
\alias{print.quickSentiment_eval}
\title{Print quickSentiment Evaluation Results}
\usage{
\method{print}{quickSentiment_eval}(x, ...)
}
\arguments{
\item{x}{An object of class `quickSentiment_eval`.}

\item{...}{Further arguments passed to or from other methods.}
}
\description{
Print quickSentiment Evaluation Results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nb.R
\name{nb_model}
\alias{nb_model}
\title{Multinomial Naive Bayes for Text Classification}
\usage{
nb_model(train_vectorized, Y, test_vectorized, parallel = FALSE, tune = FALSE)
}
\arguments{
\item{train_vectorized}{The training feature matrix (e.g., a `dfm` from quanteda).
This should be a sparse matrix.}

\item{Y}{The response variable for the training set. Should be a factor for
classification.}

\item{test_vectorized}{The test feature matrix, which must have the same
features as `train_vectorized`}

\item{parallel}{Logical}

\item{tune}{Logical. If TRUE, tests different Laplace smoothing values.}
}
\value{
A list containing four elements:
  \item{pred}{A vector of class predictions for the test set.}
  \item{probs}{A matrix of predicted probabilities.}
  \item{model}{The final, trained `naivebayes` model object.}
  \item{best_lambda}{Placeholder (NULL) for pipeline consistency.}
}
\description{
Multinomial Naive Bayes for Text Classification
}
\examples{
# 1. Create dummy numeric matrices with BOTH row and column names
train_matrix <- matrix(
  as.numeric(sample(0:5, 100, replace = TRUE)),
  nrow = 10, ncol = 10,
  dimnames = list(paste0("doc", 1:10), paste0("word", 1:10))
)

test_matrix <- matrix(
  as.numeric(sample(0:5, 50, replace = TRUE)),
  nrow = 5, ncol = 10,
  dimnames = list(paste0("doc", 1:5), paste0("word", 1:10))
)

# 2. Create dummy target variable
y_train <- factor(sample(c("P", "N"), 10, replace = TRUE))

# 3. Run model
model_results <- nb_model(train_matrix, y_train, test_matrix)
print(model_results$pred)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BOW.R
\name{BOW_train}
\alias{BOW_train}
\title{Train a Bag-of-Words Model}
\usage{
BOW_train(doc, weighting_scheme = "bow", ngram_size = 1)
}
\arguments{
\item{doc}{A character vector of documents to be processed.}

\item{weighting_scheme}{A string specifying the weighting to apply.
Defaults to \code{"bag_of_words"}.
\itemize{
  \item \code{"bag_of_words"} (Alias: \code{"bow"}) - Standard count of words.
  \item \code{"term_frequency"} (Alias: \code{"tf"}) - Normalized counts (frequency relative to document length).
  \item \code{"tfidf"} (Alias: \code{"tf-idf"}) - Term Frequency-Inverse Document Frequency.
  \item \code{"binary"} - Presence/Absence (1/0).
}}

\item{ngram_size}{An integer specifying the maximum n-gram size. For example,
`ngram_size = 1` will create unigrams only; `ngram_size = 2` will create unigrams and bigrams. Defaults to 1.}
}
\value{
An object of class \code{"qs_bow_fit"} containing:
  \itemize{
    \item \code{dfm_template}: a quanteda \code{dfm} template
    \item \code{weighting_scheme}: the weighting used
    \item \code{ngram_size}: the n-gram size used
  }#'
}
\description{
Train a Bag-of-Words Model
}
\examples{
txt <- c("text one", "text two text")
fit <- BOW_train(txt, weighting_scheme = "bow")
fit$dfm_template

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logit.R
\name{logit_model}
\alias{logit_model}
\title{Train a Regularized Logistic Regression Model using glmnet}
\usage{
logit_model(
  train_vectorized,
  Y,
  test_vectorized,
  parallel = FALSE,
  tune = FALSE
)
}
\arguments{
\item{train_vectorized}{The training feature matrix (e.g., a `dfm` from quanteda).
This should be a sparse matrix.}

\item{Y}{The response variable for the training set. Should be a factor for
classification.}

\item{test_vectorized}{The test feature matrix, which must have the same
features as `train_vectorized`.}

\item{parallel}{Logical}

\item{tune}{Logical}
}
\value{
A list containing two elements:
  \item{pred}{A vector of class predictions for the test set.}
  \item{probs}{A matrix of predicted probabilities.}
  \item{model}{The final, trained `cv.glmnet` model object.}
  \item{best_lambda}{The optimal lambda value found during cross-validation.}
}
\description{
This function trains a logistic regression model using Lasso regularization
via the glmnet package. It uses cross-validation to automatically find the
optimal regularization strength (lambda).
}
\examples{
\dontrun{
# Create dummy vectorized training and test data
train_matrix <- matrix(runif(100), nrow = 10, ncol = 10)
test_matrix <- matrix(runif(50), nrow = 5, ncol = 10)

# Provide column names (vocabulary) required by glmnet
colnames(train_matrix) <- paste0("word", 1:10)
colnames(test_matrix) <- paste0("word", 1:10)

y_train <- factor(sample(c("P", "N"), 10, replace = TRUE))

# Run logistic regression model (glmnet)
model_results <- logit_model(train_matrix, y_train, test_matrix)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{pipeline}
\alias{pipeline}
\title{Run a Full Text Classification Pipeline on Preprocessed Text}
\usage{
pipeline(
  vect_method,
  model_name,
  text_vector,
  sentiment_vector,
  n_gram = 1,
  tune = FALSE,
  parallel = FALSE
)
}
\arguments{
\item{vect_method}{A string specifying the vectorization method.
Defaults to \code{"bag_of_words"}.
\itemize{
  \item \code{"bag_of_words"} (Alias: \code{"bow"}) - Standard count of words.
  \item \code{"term_frequency"} (Alias: \code{"tf"}) - Normalized counts.
  \item \code{"tfidf"} (Alias: \code{"tf-idf"}) - Term Frequency-Inverse Document Frequency.
  \item \code{"binary"} - Presence/Absence (1/0).
}}

\item{model_name}{A string specifying the model to train.
Defaults to \code{"logistic_regression"}.
\itemize{
  \item \code{"random_forest"} (Alias: \code{"rf"})
  \item \code{"xgboost"} (Alias: \code{"xgb"})
  \item \code{"logistic_regression"} (Alias: \code{"logit"}, \code{"glm"})
}}

\item{text_vector}{A character vector containing the **preprocessed** text.}

\item{sentiment_vector}{A vector or factor containing the target labels (e.g., ratings).}

\item{n_gram}{The n-gram size to use for BoW/TF-IDF. Defaults to 1.}

\item{tune}{Logical. If TRUE, the pipeline will perform hyperparameter tuning
for the selected model. Defaults to FALSE. [NEW]}

\item{parallel}{If TRUE, runs model training in parallel. Default FALSE.}
}
\value{
A list containing the trained model object, the DFM template,
  class levels, and a comprehensive evaluation report.
}
\description{
This function takes a data frame with pre-cleaned text and handles the
data splitting, vectorization, model training, and evaluation.
}
\examples{
df <- data.frame(
  text = c("good product", "excellent", "loved it", "great quality",
           "bad service", "terrible", "hated it", "awful experience",
           "not good", "very bad", "fantastic", "wonderful"),
  y = c("P", "P", "P", "P", "N", "N", "N", "N", "N", "N", "P", "P")
)


out <- pipeline("bow", "naive_bayes",  text_vector = df$text, sentiment_vector = df$y)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pre_process.R
\docType{data}
\name{qs_negations}
\alias{qs_negations}
\title{Standard Negation Words for Sentiment Analysis}
\format{
An object of class \code{character} of length 25.
}
\usage{
qs_negations
}
\description{
A character vector of 25 common negation words. These words are automatically
protected by the \code{\link{pre_process}} function when \code{retain_negations = TRUE}
to prevent standard stopword lists from destroying sentiment polarity.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BOW.R
\name{BOW_test}
\alias{BOW_test}
\title{Transform New Text into a Document-Feature Matrix}
\usage{
BOW_test(doc, fit)
}
\arguments{
\item{doc}{A character vector of new documents to be processed.}

\item{fit}{A fitted BoW object returned by \code{BOW_train()}.}
}
\value{
A quanteda \code{dfm} aligned to the training features.
}
\description{
This function takes a character vector of new documents and transforms it
into a DFM that has the exact same features as a pre-fitted training DFM,
ensuring consistency for prediction.
}
\examples{
train_txt <- c("apple orange banana", "apple apple")
fit <- BOW_train(train_txt, weighting_scheme = "bow")
new_txt <- c("banana pear", "orange apple")
test_dfm <- BOW_test(new_txt, fit)
test_dfm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evaluate_performance.R
\name{evaluate_performance}
\alias{evaluate_performance}
\title{Evaluate Model Performance (ROC and Precision-Recall)}
\usage{
evaluate_performance(predicted_probs, actual_classes, positive_label)
}
\arguments{
\item{predicted_probs}{Numeric vector of predicted probabilities for the positive class.}

\item{actual_classes}{Factor or character vector of the actual true labels.}

\item{positive_label}{Character string. The target class you want to evaluate.}
}
\value{
An object of class `quickSentiment_eval`, which is a list containing the following metrics:
  \item{target_class}{Character. The specific positive label used for the evaluation.}
  \item{auc_roc}{Numeric. The Area Under the Receiver Operating Characteristic curve.}
  \item{best_threshold_roc}{Numeric. The optimal probability threshold that maximizes Youden's J statistic.}
  \item{auc_pr}{Numeric. The Area Under the Precision-Recall curve.}
  \item{best_threshold_pr}{Numeric. The probability threshold that maximizes the F1-Score.}
  \item{accuracy_at_best}{Numeric. The overall accuracy of the model if `best_threshold_pr` is applied.}
  \item{roc}{S3 object containing the ROC curve data and metrics.}
  \item{prc}{S3 object containing the Precision-Recall curve data and metrics.}
  \item{threshold_summary}{A data frame summarizing Accuracy, Precision, Recall, and F1 at 0.1 threshold increments.}
}
\description{
Evaluate Model Performance (ROC and Precision-Recall)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pre_process.R
\name{pre_process}
\alias{pre_process}
\title{Preprocess a Vector of Text Documents}
\usage{
pre_process(
  doc_vector,
  remove_brackets = TRUE,
  remove_urls = TRUE,
  remove_html = TRUE,
  remove_nums = FALSE,
  remove_emojis_flag = TRUE,
  to_lowercase = TRUE,
  remove_punct = TRUE,
  remove_stop_words = TRUE,
  custom_stop_words = NULL,
  keep_words = NULL,
  lemmatize = TRUE,
  retain_negations = TRUE
)
}
\arguments{
\item{doc_vector}{A character vector where each element is a document.}

\item{remove_brackets}{A logical value indicating whether to remove text in square brackets.}

\item{remove_urls}{A logical value indicating whether to remove URLs and email addresses.}

\item{remove_html}{A logical value indicating whether to remove HTML tags.}

\item{remove_nums}{A logical value indicating whether to remove numbers.}

\item{remove_emojis_flag}{A logical value indicating whether to remove common emojis.}

\item{to_lowercase}{A logical value indicating whether to convert text to lowercase.}

\item{remove_punct}{A logical value indicating whether to remove punctuation.}

\item{remove_stop_words}{A logical value indicating whether to remove English stopwords.}

\item{custom_stop_words}{A character vector of additional custom words to remove (e.g., c("rt", "via")). Default is NULL.}

\item{keep_words}{A character vector of words to protect from deletion (e.g., c("no", "not", "nor")). Default is NULL.}

\item{lemmatize}{A logical value indicating whether to lemmatize words to their dictionary form.}

\item{retain_negations}{Logical. If \code{TRUE} (the default), automatically protects common negation words (e.g., "not", "no", "never") from being deleted by the standard stopword list to preserve sentiment context.}
}
\value{
A character vector of the cleaned and preprocessed text.
}
\description{
This function provides a comprehensive and configurable pipeline for cleaning
raw text data. It handles a variety of common preprocessing steps including
removing URLs and HTML, lowercasing, stopword removal, and lemmatization.
}
\examples{
raw_text <- c(
  "This is a <b>test</b>! Visit https://example.com",
  "Email me at test.user@example.org [important]"
)

# Basic preprocessing with defaults
clean_text <- pre_process(raw_text)
print(clean_text)

# Keep punctuation and stopwords
clean_text_no_stop <- pre_process(
  raw_text,
  remove_stop_words = FALSE,
  remove_punct = FALSE
)
print(clean_text_no_stop)
}
